<!--suppress HtmlUnknownTarget -->
<!DOCTYPE html>
<html>
  <head>
    <!--<meta http-equiv="refresh" content="5">-->
    <title>boofuzz Fuzz Control</title>
    <link rel="stylesheet" type="text/css" href="/static/css/boofuzz.css">
    <link rel="shortcut icon" href="{{ url_for('static', filename='favicon-32x32.ico') }}">
    <link rel="icon" type="image/svg+xml" href="{{ url_for('static', filename='favicon.svg') }}" sizes="any">
  </head>
  <body>
  {% block body %} {% endblock %}
  </body>
</html>
