from . import ifuzz_logger

# IFuzzLoggerBackend is identical to IFuzzLogger at this point, so just copy:
IFuzzLoggerBackend = ifuzz_logger.IFuzzLogger
"""
Interface for backends for FuzzLogger. Currently identical to IFuzzLogger.
"""
